# -*- coding:utf-8 -*

import numpy as np

class BinaryCrossEntropy:
    '''
    loss = -ylog(py) - (1-y)log(1-py)
    '''

    def __init__(self, epsilon=1e-8):
        self.epsilon = epsilon
        self.s = 0.0
        self.count = 0

    # y: real value
    # py: predict value
    def calculate(self, y, py):
        self.s += -y*np.log(py+self.epsilon) - (1-y)*np.log(1-py+self.epsilon)
        self.__devirate(y, py)
        self.count += 1
        
    def loss(self):
        return self.s/self.count

    def error(self):
        return self.e

    def __devirate(self, y, py):
        self.e = -y/(py+self.epsilon) - (1-y)/(1-py+self.epsilon)


class CategoryCrossEntropy:

    '''
    loss = -ylog(py)
    '''
    def __init__(self, epsilon=1e-8):
        self.epsilon = epsilon
        self.count = 0
        self.s = 0.0

    def calculate(self, y, py):
        self.s += -y*np.log(py+self.epsilon)
        self.__derivate(y, py)
        self.count += 1

    def loss(self):
        return np.sum(self.s/self.count)

    def error(self):
        return self.e

    def __derivate(self, y, py):
        self.e = -y/(py+self.epsilon)


# -*- coding:utf-8 -*-

import numpy as np

'''
--------------------------------------------------- dense layer --------------------------------------------------- 
'''

class Dense:

    # args: input_dim
    def __init__(self, output_dim, **args):
        self.out_dim = output_dim
        self.in_dim = None
        if args and args['input_dim']:
            self.in_dim = args['input_dim']
            self.w = np.random.randn(self.in_dim, self.out_dim) - 0.5
        self.b = np.random.randn(1, self.out_dim)
        self.w_optimizer, self.b_optimizer = None, None

    def is_activation(self):
        return False

    def set_input_dim(self, input_dim):
        self.in_dim = input_dim
        self.w = np.random.randn(self.in_dim, self.out_dim) - 0.5

    def input_dim(self):
        return self.in_dim

    def output_dim(self):
        return self.out_dim

    def weight(self):
        return self.w
        
    # x input
    # 矩阵相乘
    def calculate(self, x):
        self.x = x
        self.y = np.dot(self.x, self.w) + self.b

    # 返回计算值
    def output(self):
        return self.y

    # 求导 计算误差
    # error
    def derivate(self, error):
        self.ew = np.dot(self.x.T, error)
        self.eb = error
        return self.ew, self.eb

    def update(self, error_weight, error_bias):
        self.b += error_bias
        self.w += error_weight

    def set_optimizer(self, weight_optimizer_function, bias_optimizer_function):
        self.w_optimizer = weight_optimizer_function
        self.b_optimizer = bias_optimizer_function

    def weight_optimizer(self):
        return self.w_optimizer
    
    def bias_optimizer(self):
        return self.b_optimizer


'''
--------------------------------------------------- activation function --------------------------------------------------- 
'''

# -------------------------------- Relu --------------------------------

class Relu:

    def __init__(self, a):
        a = max(0, a)
        self.a = a

    def is_activation(self):
        return True

    def calculate(self, x):
        self.x = x
        self.y = np.maximum(self.x, self.a * self.x)

    def output(self):
        return self.y

    def derivate(self, e):
        self.ex = np.where(self.y>0, e, self.a*e)
        return self.ex



# -------------------------------- Sigmoid --------------------------------

class Sigmoid:

    def __init__(self):
        pass

    def is_activation(self):
        return True

    def calculate(self, x):
        self.x = x
        self.y = 1.0/(1.0+np.exp(-x))

    def output(self):
        return self.y

    def derivate(self, e):
        self.ex = e * (self.y * (1-self.y))
        return self.ex



# -------------------------------- Tanh --------------------------------

class Tanh:

    def __init__(self):
        pass

    def is_activation(self):
        return True

    def calculate(self, x):
        self.x = x
        self.y = np.tanh(x)
    
    def output(self):
        return self.y

    def derivate(self, e):
        self.ex = e * (1 - self.y * self.y)
        return self.ex

# -------------------------------- dropout --------------------------------

class Dropout:

    # drop 丢弃概率
    def __init__(self, drop):
        self.drop = drop
        self.retain_prob = 1. - drop

    def is_activation(self):
        return True

    def calculate(self, x):
        self.x = x
        self.r = np.random.binomial(n=1, p=self.retain_prob, size=self.x.shape)  #  r = 0|1
        self.y = self.x * self.r / self.retain_prob

    def output(self):
        return self.y

    def derivate(self, e):
        self.ex = e * (self.r / self.retain_prob)
        return self.ex


# -------------------------------- softmax --------------------------------

class Softmax:

    def __init__(self):
        pass

    def is_activation(self):
        return True

    def calculate(self, x):
        m = np.max(x)
        s = np.exp(x-m)
        self.y = s/np.sum(s)

    def output(self):
        return self.y
    
    def derivate(self, e):
        ex = np.ones_like(self.y)
        for i in range(len(self.y[0])):
            a=np.ones_like(self.y) * -self.y[0][i]
            a *= self.y
            a[0][i] += self.y[0][i]
            ex[0][i] *= np.sum(e*a)
        self.ex = ex
        return self.ex


#  -*- coding:utf-8 -*-

import numpy as np

class Adagrad:

    def __init__(self, learning_rate=0.01, epsilon=0.00001):
        self.learning_rate = learning_rate
        self.epsilon = epsilon
        self.prev_n = None

    def __init_n(self, dim):
        self.prev_n = np.zeros(dim)

    def grad(self, grad):
        if self.prev_n is None:
            self.__init_n(grad.shape)
        self.prev_n = self.prev_n + grad*grad
        return -self.learning_rate*grad/(np.sqrt(self.prev_n+self.epsilon))


#-*- coding:utf-8 -*-
 
#############################################
# File Name: setup.py
# Author: Frank Lee
# Mail: frankleecsz@gmail.com
# Created Time: 2020-01-18
#############################################
 
 
from setuptools import setup, find_packages
 
setup(
  name = "teny",
  version = "0.1.3",
  keywords = ["pip", "dnn", "DeepLearning", "MachineLearning"],
  description = "dnn framework",
  long_description = "simple deep learning framework in numpy for demo.",
  license = "MIT Licence",
 
  url = "https://github.com/FrankLeeC/MachineLearning/teny",
  author = "FrankLee",
  author_email = "frankleecsz@gmail.com",
 
  packages = find_packages(),
  include_package_data = True,
  platforms = "any",
  install_requires = ['numpy']
)


# -*- coding:utf-8 -*-
import numpy as np
import logging
import sys
import copy
 
class NetWork:

    def __init__(self, name):
        self.name = name
        self.neurals = []
        self.last_in_dim = None
        self.__init_log()
        self.init_opt = False

    def __init_log(self):
        self.logger = logging.getLogger('log[' + self.name + ']')
        self.log_formatter = logging.Formatter('%(asctime)s %(levelname)-5s: %(message)s')
        log_handler = logging.StreamHandler(sys.stdout)
        log_handler.setFormatter(self.log_formatter)
        self.logger.addHandler(log_handler)
        self.logger.setLevel(logging.INFO)
        

    def add(self, neural):
        if (not neural.is_activation()) and (neural.input_dim() is None):
            neural.set_input_dim(self.last_in_dim)
        if not neural.is_activation():  # dense
            self.last_in_dim = neural.output_dim()
        self.neurals.append(neural)

    def loss(self, f):
        self.loss_function = f

    def optimizer(self, opt):
        self.optimizer_function = opt

    def add_logger_handler(self, log_handler):
        log_handler.setFormatter(self.log_formatter)
        self.logger.addHandler(log_handler)

    def __output(self, s):
        self.logger.info('['+self.name+'] '+s)

    def __init_optimizer(self):
        for l in self.neurals:
            if not l.is_activation():
                l.set_optimizer(copy.deepcopy(self.optimizer_function), copy.deepcopy(self.optimizer_function))
        
    # iteration 全数据量的训练次数
    # batch_size 每个 epoch 数据量
    # 所有 epoch 训练一次 算 一次 iteration
    def train(self, x, y, batch_size, iteration=1000):
        if not self.init_opt:
            self.init_opt = True
            self.__init_optimizer()
        self.batch_size = batch_size
        epochs = int(len(y) / batch_size)
        if len(y) % batch_size != 0:
            epochs += 1
        for _ in range(iteration):
            for e in range(epochs):
                start = e * batch_size
                end = (e+1) * batch_size
                train_x, train_y = x[start:end], y[start:end]
                self.__batch(train_x, train_y)
            self.__output('loss: ' + str(self.__loss()))
        
    
    def __batch(self, x, y):
        self.error_weight, self.error_bias = [], []

        # init
        for l in self.neurals:
            a, b = 0, 0
            if not l.is_activation():
                a, b = l.input_dim(), l.output_dim()
            self.error_weight.append(np.zeros([a, b]))
            self.error_bias.append(np.zeros([1,b]))

        # 计算样本
        for e, l in zip(x, y):
            p = self.__forward(e)
            self.__calculate_loss(l, p)
            self.__backprocess()
        
        # 计算平均误差
        for i in range(len(self.neurals)):
            if not self.neurals[i].is_activation():
                self.error_weight[i]/self.batch_size
                self.error_bias[i]/self.batch_size

        # 更新
        for i, l in enumerate(self.neurals):
            if not l.is_activation():
                ew , eb = l.weight_optimizer().grad(self.error_weight[i]), l.bias_optimizer().grad(self.error_bias[i])
                ew = np.clip(ew, -1.0, 1.0)
                eb = np.clip(eb, -1.0, 1.0)
                l.update(ew, eb)

        

    # 每计算一个样本，执行一次
    def __forward(self, x):
        p = x
        for _, l in enumerate(self.neurals):
            l.calculate(p)
            p = l.output()
        return p

    def __calculate_loss(self, y, py):
        self.loss_function.calculate(y, py)

    def __loss(self):
        return self.loss_function.loss()

    def __error(self):
        return self.loss_function.error()

    # 每计算一个样本，执行一次
    def __backprocess(self):
        der_error = self.__error()
        for i in range(len(self.neurals)):
            j = len(self.neurals)-i-1
            if self.neurals[j].is_activation():
                der_error = self.neurals[j].derivate(der_error)
            else:
                a, b = self.neurals[j].derivate(der_error)
                self.error_weight[j] += np.reshape(a, self.error_weight[j].shape)
                self.error_bias[j] += np.reshape(b, self.error_bias[j].shape)
                if j > 0:
                    der_error = np.dot(der_error, self.neurals[j].weight().T)
                

    # x array
    def predict(self, x):
        return np.array([self.__forward(a) for a in x])
